
#include "local.h"
#include <stdio.h>
#include <math.h>
#include "son.h"

/*************************************************************/
/* Routines for manipulating circuit theory parameters.      */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr24may91 */ 

/* MKS units used at all times */
#include "em.typ"  /* Data structure definitions */

xfrm_ports(ports)  port_list *ports;
/*************************************************************/
/* Merge port subsections together.                          */
/* (C) COPYRIGHT 1991 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr02may91 */ 
{ port_typ *iport;
  int max,found,i,j;

  /*Transfer user specified port indices to line_nu*/
  /*This info is required for comments in output file*/
  /*Also set sign if port index is negative and make index positive*/
  for (iport=ports->port0; iport<ports->port0+ports->n; iport++)
  { iport->line_nu = iport->index;
    if ( iport->index<0 )
    { iport->sign = -1;
      iport->index = -iport->index;
    }
    else iport->sign = 1;
  }

  /*Set up port indices, first find max index*/
  max = 0;
  for (iport=ports->port0; iport<ports->port0+ports->n; iport++)
    if ( iport->index>max ) max = iport->index;
  /*Set_up new indices consequitively starting with zero*/
  for (i=0,j=0; i<=max; i++)
  { found = 0;
    for (iport=ports->port0; iport<ports->port0+ports->n; iport++)
      if ( iport->index==i ) { iport->index = j;  found = 1; }
    if (found) j++;
  } 
  ports->m = j;  /*Number of actual ports, some of the ports->n ports might*/
                 /*be shorted together (i.e., given same index)*/

} /*xfrm_ports*/

set_port_norm(ports,An)  port_list *ports;  An_info *An;
/*************************************************************/
/* Set up port nomalizing information.                       */
/* (C) COPYRIGHT 1991 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr21jun91 */
{ int i;
  port_typ *port;

  if (An->rp || An->xp || An->lp || An->cp)
  { announce_bug("set_port_norm"); }

  An->rp = (Rtyp *)cnew( ports->m,sizeof(Rtyp) );
  An->xp = (Rtyp *)cnew( ports->m,sizeof(Rtyp) );
  An->lp = (Rtyp *)cnew( ports->m,sizeof(Rtyp) );
  An->cp = (Rtyp *)cnew( ports->m,sizeof(Rtyp) );

  for (i=0; i<ports->m; i++)
    for (port=ports->port0; port<ports->port0+ports->n; port++)
      if (port->index == i)
      { *(An->rp+i) = port->rp;
        *(An->xp+i) = port->xp;
        *(An->lp+i) = port->lp;
        *(An->cp+i) = port->cp;
        break;
      }
} /*set_port_norm*/

abcdz(R)  Mat_ptr R;
/*************************************************************/
/* Convert between Z and ABCD parameters.                    */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr04sep87 */
/* jcr24jan88 -- any 2-N port*/
{ int i,j,n;
  Vec_ptr Temp;

  /*R is 2n x 2n.  No check for error of odd matrix size*/
  n = R->Num_col / 2;
  /*Swap first n rows with last n rows. */
  for (i=0; i<n; i++)
  { Temp = R->Row[i];
    R->Row[i] = R->Row[i+n];
    R->Row[i+n] = Temp;
  }
  /*Do a partial inversion*/
  Mat_partinv(n,R,0);
  /*Swap the rows back*/
  for (i=0; i<n; i++)
  { Temp = R->Row[i];
    R->Row[i] = R->Row[i+n];
    R->Row[i+n] = Temp;
  }
  /*Change the sign of the last n columns*/
  for (i=0; i<2*n; i++) for (j=n; j<2*n; j++)
    Mat_put(i,j, -Mat_get_re(i,j,R), -Mat_get_im(i,j,R), R);
} /*abcdz*/

sabcd(R)  Mat_ptr R;
/*************************************************************/
/* Convert S parameters to ABCD parameters.                  */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr11feb88 */ 
/* This conversion will not work for a zero length thru */
{ stoz(R);
  abcdz(R);
} /*sabcd*/

abcds(R)  Mat_ptr R;
/*************************************************************/
/* Convert between from ABCD to S parameters.                */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr11feb88 */ 
/* Does not get intermediate results in terms of Z or Y parameters
   so that a zero length thru may be processed. */
{ int i,n;
  Vec_ptr Tv;  /*Temporary vector*/
  Mat_ptr Tm;  /*Temporary matrix*/

  /*R is 2n x 2n.  No check for error of odd matrix size*/
  n = R->Num_col / 2;

  /*Convert voltage and current to a and b for each port...

      a = 0.5 * ( V + Zo*i )
      b = 0.5 * ( V - Zo*i )

  */

  Tm = Mat_open(R->Num_col,R->Num_col);
  /* Convert input variables into a and b by manipulating rows */
  Tv = Vec_open(R->Num_col);
  for (i=0; i<n; i++)
  { Vec_smult(50.0,0.0,R->Row[i+n],Tv);
    Vec_add(R->Row[i],Tv,Tm->Row[i]);  /* 2*a */
    Vec_sub(R->Row[i],Tv,Tm->Row[i+n]);  /* 2*b */
  }
  /* now (a1,b1) = Tm * (v2,i2) */
  /*Put the output variables on the input side*/
  Mat_inv(Tm,0);
  /*Now convert the output variables to a and b*/
  /*Remember that the last n variables are -i */
  for (i=0; i<n; i++)
  { Vec_smult(-50.0,0.0,Tm->Row[i+n],Tv);
    Vec_add(Tm->Row[i],Tv,R->Row[i+n]);  /* 2*a, last n */
    Vec_sub(Tm->Row[i],Tv,R->Row[i]);  /* 2*b, 1st n */
  }
  /*Now (b2,a2) = R * (a1,b1) */
  /*Put the output variables back on the output side*/
  Mat_inv(R,0);
  Vec_close(Tv);
  Mat_close(Tm);

  /*Do a partial inversion*/
  Mat_partinv(n,R,0);
  /*Now (b2,b1) = R * (a1,a2) */
  /*Swap the rows so input b on top*/
  for (i=0; i<n; i++)
  { Tv = R->Row[i];
    R->Row[i] = R->Row[i+n];
    R->Row[i+n] = Tv;
  }
  /*Now b = S*a */
} /*abcds*/

stoz(R0)  Mat_ptr R0;
/*************************************************************/
/* Convert from S to Z parameters.                           */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr11feb88 */ 
{ Mat_ptr R,D;
  int i;

  /*Put R0 into R and manipulate R because Mat_mult changes 
    where it's answer matrix is stored */
  R = Mat_open(R0->Num_row,R0->Num_col);
  Mat_move(R0,R);
  D = Mat_open(R->Num_row,R->Num_col);

  /*   Z = Zo * (1 + S) / (1 - S)  */

  Mat_move(R,D);
  for (i=0; i<R->Num_col; i++) 
  { Mat_put_re( i,i, Mat_get_re(i,i,R)+1.0, R );
    Mat_put_re( i,i, Mat_get_re(i,i,D)-1.0, D );
  }
  Mat_inv(D,0);
  Mat_mult(D,R,&R);
  Mat_smult(-50.0,0.0,R,R0);
  Mat_close(R);   Mat_close(D);   
} /*stoz*/

ztos(R0)  Mat_ptr R0;
/*************************************************************/
/* Convert from Z to S parameters.                           */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr11feb88 */ 
{ Mat_ptr R,D;
  int i;

  /*Put R0 into R and manipulate R because Mat_mult changes 
    where it's answer matrix is stored */
  R = Mat_open(R0->Num_row,R0->Num_col);
  Mat_move(R0,R);
  D = Mat_open(R->Num_row,R->Num_col);

  /*   S = (Z - Zo) / (Z + Zo)  */

  Mat_move(R,D);
  for (i=0; i<R->Num_col; i++) 
  { Mat_put_re( i,i, Mat_get_re(i,i,R)-50.0, R );
    Mat_put_re( i,i, Mat_get_re(i,i,D)+50.0, D );
  }
  Mat_inv(D,0);
  Mat_mult(D,R,&R);
  Mat_move(R,R0);
  Mat_close(R);   Mat_close(D);   
} /*stoz*/

stoy(R0)  Mat_ptr R0;
/*************************************************************/
/* Convert from S to Y parameters.                           */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr11feb88 */ 
{ Mat_ptr R,D;
  int i;

  /*Put R0 into R and manipulate R because Mat_mult changes 
    where it's answer matrix is stored */
  R = Mat_open(R0->Num_row,R0->Num_col);
  Mat_move(R0,R);
  D = Mat_open(R->Num_row,R->Num_col);

  /*   Y = Yo * (1 - S) / (1 + S)  */

  Mat_move(R,D);
  for (i=0; i<R->Num_col; i++) 
  { Mat_put_re( i,i, Mat_get_re(i,i,R)-1.0, R );
    Mat_put_re( i,i, Mat_get_re(i,i,D)+1.0, D );
  }
  Mat_inv(D,0);
  Mat_mult(D,R,&R);
  Mat_smult(-0.02,0.0,R,R0);
  Mat_close(R);   Mat_close(D);   
} /*stoy*/

ytos(R0)  Mat_ptr R0;
/*************************************************************/
/* Convert from Y to S parameters.                           */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr11feb88 */ 
{ Mat_ptr R,D;
  int i;

  /*Put R0 into R and manipulate R because Mat_mult changes 
    where it's answer matrix is stored */
  R = Mat_open(R0->Num_row,R0->Num_col);
  Mat_move(R0,R);
  D = Mat_open(R->Num_row,R->Num_col);

  /*   S = (Yo - Y) / (Yo + Y)  */

  Mat_move(R,D);
  for (i=0; i<R->Num_col; i++) 
  { Mat_put_re( i,i, Mat_get_re(i,i,R)-0.02, R );
    Mat_put_re( i,i, Mat_get_re(i,i,D)+0.02, D );
  }
  Mat_inv(D,0);
  Mat_mult(D,R,&R);
  Mat_smult(-1.0,0.0,R,R0);
  Mat_close(R);   Mat_close(D);   
} /*ytos*/

calc_z(R,X,L,C,F,rep,imp)  Rtyp R,X,L,C,F,*rep,*imp;
/*************************************************************/
/* Calculate the impedance of a simple network.              */
/* (C) COPYRIGHT 1991 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr24may91 */
{ complex y,z;

  /*    F is frequency in MHz */

  /*    O---*---R(OHMS)----X(OHMS)----L(HENRIES)---
   *        |                                     |
   *      C(FARADS)                               |
   *        |                                     |
   *       GND                                   GND
   */

  *rep = R;
  *imp = X + TWO_PI * F * 1000000.0 * L;
  if (fabs(C)>TINY)
  { z.Re = *rep;
    z.Im = *imp;
    y.Re = 0.0;
    y.Im = TWO_PI * F * 1000000.0 * C;
    Cmult(&y,&z,&y);
    y.Re += 1.0;
    Cdiv(&z,&y,&z);
    *rep = z.Re;
    *imp = z.Im;
  }
  return;
} /*calc_z*/

renorm_s(S,rp,xp,lp,cp,F)  Mat_ptr S;  Rtyp *rp,*xp,*lp,*cp,F;
/*************************************************************/
/* Renormalize S parameters to new impedance level.          */
/* (C) COPYRIGHT 1991 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr24may91 */
{ int i,j,z_param;
  complex z0,z1;
  Rtyp rval,xval,lval,cval;

  z_param = 0;
  for (i=0; i<S->Num_row; i++)
  { rval = (rp) ? *(rp+i) : 50.0;
    xval = (xp) ? *(xp+i) : 0.0;
    lval = (lp) ? *(lp+i) : 0.0;
    cval = (cp) ? *(cp+i) : 0.0;
    calc_z(rval,xval,lval,cval,F,&(z1.Re),&(z1.Im));
    if ( fabs(z1.Im)>TINY || fabs(z1.Re-50.0)>1.0e-5 )  /*renormlize*/
    { if (!z_param) { stoz(S);  z_param = 1; }
      z1.Re *= 0.02;  z1.Im *= 0.02;  /*Divide by 50*/
      Csqrt(&z1,&z1);
      for (j=0; j<S->Num_row; j++)
      { z0.Re = Mat_get_re(i,j,S);
        z0.Im = Mat_get_im(i,j,S);
        Cdiv(&z0,&z1,&z0);
        Mat_put(i,j,z0.Re,z0.Im,S);
        z0.Re = Mat_get_re(j,i,S);
        z0.Im = Mat_get_im(j,i,S);
        Cdiv(&z0,&z1,&z0);
        Mat_put(j,i,z0.Re,z0.Im,S);
      }
    }
  }
  if (z_param) ztos(S);
  return;
} /*renorm_s*/

Mat_ptr shrt_ports(ports,R,J,Rp,Jp)  Mat_ptr R,J,*Rp,*Jp;  port_list *ports;
/*************************************************************/
/* Re-order (& shrt together) ports based on index.          */
/* (C) COPYRIGHT 1991 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr02may91 */ 
/* J is current distribution, if none, just pass NULL*/
{ Mat_ptr T1,T2;  /*Merged ports matrices*/
  Mat_ptr J1,J2;  /*Temps for J, current distribution*/
  port_typ *iport,*jport;
  int i,j,row,col;
  int *in_sto;  /*Store which indices are active*/
  Rtyp re,im;

  if (ports->n != R->Num_row)
    announce_bug("shrt_ports(0)");

  /*Check for any short or open circuit S-Parameters*/
  /*Such can cause Y or Z param to blow up*/
  /*If found, make them slightly different from short or open*/
  for (i=0; i<R->Num_row; i++)
    for (j=i; j<R->Num_col; j++)
    { re = Mat_get_re(i,j,R);
      im = Mat_get_im(i,j,R);
      if ( fabs(1.0-re) + fabs(im) < 1.0E-6 )  /*Short circuit*/
      { Mat_put_im(i,j,1.0E-6,R);
        if ( i != j )
          Mat_put_im(j,i,1.0E-6,R);
      }
      if ( fabs(1.0+re) + fabs(im) < 1.0E-6 )  /*Open circuit*/
      { Mat_put_im(i,j,1.0E-6,R);
        if ( i != j )
          Mat_put_im(j,i,1.0E-6,R);
      }
    }

  J1 = J2 = T1 = T2 = NULL;

  /*Positive sign rows even, neg rows are odd*/
  T1 = Mat_open(2*ports->m,2*ports->m);
  if (J) J1 = Mat_open(2*ports->m,J->Num_col);
  /*List of which rows see data*/
  in_sto = (int *)cnew( 2 * ports->m, sizeof(int) );

  /*Initialize to unit matrix*/
  for (i=0; i<T1->Num_row; i++) Mat_put_re(i,i,1.0,T1);

  /*Zero out rows that will accumulate data and set in_sto*/
  for (i=0; i<ports->n; i++)
  { iport = ports->port0 + i;
    row = 2 * iport->index;
    if (iport->sign<0) row++;
    Mat_put_re(row,row,0.0,T1);
    *(in_sto+row) = 1;
  }

  /*Convert to Y-param*/
  stoy(R);

  /*Do first stage merge*/
  for (i=0; i<ports->n; i++)
  { iport = ports->port0 + i;
    row = 2 * iport->index;
    if (iport->sign<0) row++;
    for (j=0; j<ports->n; j++)
    { jport = ports->port0 + j;
      col = 2 * jport->index;
      if (jport->sign<0) col++;
      re = Mat_get_re(i,j,R) + Mat_get_re(row,col,T1);
      im = Mat_get_im(i,j,R) + Mat_get_im(row,col,T1);
      Mat_put(row,col,re,im,T1);
    }
    if (J) Vec_add(J->Row[i],J1->Row[row],J1->Row[row]);
  }

  /*Convert to Z-parameters*/
  Mat_inv(T1,0);
  if (J) Mat_mult(T1,J1,&J1);

  /*Zero out rows/cols that didn't accumulate data*/
  for (i=0; i<T1->Num_row; i++) if ( *(in_sto+i) == 0 )
  { for (j=0; j<T1->Num_row; j++)
    { Mat_put(i,j,0.0,0.0,T1);
      Mat_put(j,i,0.0,0.0,T1);
    }
    if (J) Vec_init(0.0,0.0,J1->Row[i]);
  }

  /*Do second (and final) stage merge*/
  T2 = Mat_open(ports->m,ports->m);
  for (i=0; i<ports->m; i++) for (j=0; j<ports->m; j++)
  { re = Mat_get_re(2*i,2*j,T1) + Mat_get_re(2*i+1,2*j+1,T1);
    re = re - Mat_get_re(2*i,2*j+1,T1) - Mat_get_re(2*i+1,2*j,T1);
    im = Mat_get_im(2*i,2*j,T1) + Mat_get_im(2*i+1,2*j+1,T1);
    im = im - Mat_get_im(2*i,2*j+1,T1) - Mat_get_im(2*i+1,2*j,T1);
    Mat_put(i,j,re,im,T2);
  }
  if (J)
  { J2 = Mat_open(ports->m,J->Num_col);
    for (i=0; i<ports->m; i++)
      Vec_sub(J1->Row[2*i],J1->Row[2*i+1],J2->Row[i]);
  }

  /*Convert T2 back to S-param and J2 to Y-param*/
  if (J)
  { Mat_inv(T2,0);
    Mat_mult(T2,J2,&J2);
    ytos(T2);
  }
  else ztos(T2);

  sFree( (char *)in_sto);
  Mat_close(T1);
  if (J) Mat_close(J1);

  /*Return values*/
  *Rp = T2;
  *Jp = J2;

} /*shrt_ports*/
